//  File Pathfind_BT_33_globals.h
//  Created by: hdrake, TabbyCat Robots      2012Feb18

#ifndef _PATHFIND_BT_33_GLOBALS_H
#define _PATHFIND_BT_33_GLOBALS_H

#define BT_OPERATION_INCOMPLETE    0xff
#define BT_CONNECTION_SUCCESS         1
#define BT_CONN_ERROR                16
#define BT_MISC_STATUS_ERROR         99
#define BUTTON_PRESS_DELAY          700

#define INIT_BT_MSG_01          0x1f
#define INIT_BT_MSG_02          0x01ff
#define INIT_BT_MSG_03          0x02f
#define INIT_BT_MSG_04          0x02ff

#define NXT_0_NAME   "none"  //  Used  to initialize variables.
#define NXT_1_NAME   "WFCR"
#define NXT_8_NAME   "NCPR"
#define NXT_4_NAME   "WNYC"
#define NXT_9_NAME   "WLRN"


//  Note that the role, xger_id, conn, and queue #defines may vary
//  with routines, e.g. square dance, conga line, etc.  And there
//  will be only one type of dance per executable.  And probably only one
//  routine per executable.  Maybe someday, the routine will be a file that
//  the executable reads, and then we could string together multiple routines.

#define DNCR_ROLE_NONE                   0  //   
#define DNCR_ROLE_LEADER                 1  //  One dancer will always have
                                            //  the primary role as leader, and
                                            //  may also have a secondary role
                                            //  as first or last in line.  A
                                            //  dancer may never have both the
                                            //  role of leader and free spirit.
#define DNCR_ROLE_FREE_SPIRIT            2  //  Note that the role of FREE_SPIRIT
                                            //  may be a primary or secondary
                                            //  role for a dancer.  Having the
                                            //  role of FREE_SPIRIT gives a
                                            //  dancer some leeway in following
                                            //  the routine precisely.  e.g. The
                                            //  dancer may rotate CCW rather
                                            //  than CW or may do an extra step.
                                            //  And if the dancer's primary role
                                            //  is free spirit, then the dancer's
                                            //  steps may be a little random.
#define DNCR_ROLE_LDRS_RT               11
#define DNCR_ROLE_LDRS_OPP              12
#define DNCR_ROLE_LDRS_LFT              13

#define DNCR_ROLE_1ST_IN_LINE           21
#define DNCR_ROLE_2ND_IN_LINE           22
#define DNCR_ROLE_3RD_IN_LINE           23
#define DNCR_ROLE_4TH_IN_LINE           24

#define MAX_VALUE_FOR_ROLE              30

//  For now, there will be only 4 NXTs involved in a dance routine.
//  If, when, we give the lead dancer 2 NXTs, then ??
//  LDR is leading dancer, theleader, and will be BT master.
//  SPRT is supporting dancer.
//  Need a function that takes DNCR_ROLE_ and returns DNCR_ID.
//  The DNCR_ID is the subscript to index the dncr_info struct array.
#define LDR_DNCR_ID              0
#define SPRT_01_DNCR_ID          1
#define SPRT_02_DNCR_ID          2
#define SPRT_03_DNCR_ID          3

#define BT_CONN_00              0
#define BT_CONN_01              1
#define BT_CONN_02              2
#define BT_CONN_03              3

//  For square dance routines.
#define LDR_BT_CONN              BT_CONN_00
#define LDRS_RT_BT_CONN          BT_CONN_01
#define LDRS_OPP_BT_CONN         BT_CONN_02
#define LDRS_LFT_BT_CONN         BT_CONN_03

/*******************************************************************************
/*  For a Conga Line routine, or some other line dance, we may want
/*  the leader first in line or last.  So we might have one of the following
/*  two sets of #defines.  Remember, the leader is always the BT master.
/*
/*  #define 1ST_IN_LINE_BT_CONN   BT_CONN_01  //  Dncr using conn 1 is 1st in line
/*  #define 2ND_IN_LINE_BT_CONN   BT_CONN_02
/*  #define 3RD_IN_LINE_BT_CONN   BT_CONN_03
/*  #define LDR_BT_CONN           BT_CONN_00  //  Leader is last in line.
/*
/*  #define LDR_BT_CONN           BT_CONN_00  //  Leader is first in line.
/*  #define 2ND_IN_LINE_BT_CONN   BT_CONN_01  //  Dncr using conn 1 is 2nd in line
/*  #define 3RD_IN_LINE_BT_CONN   BT_CONN_02
/*  #define 4TH_IN_LINE_BT_CONN   BT_CONN_03
/*
/******************************************************************************/

//  For all routines
//  The leaders dncr_pri_bt_queue will always be LDR_BT_Q.
#define LDR_BT_Q                 0

//  The following _BT_Q #defines may be a dancer's dncr_pri_bt_queue,
//  dncr_sec_bt_queue, or dncr_tert_bt_queue.
#define FREE_SPIRIT_BT_Q         9

#define BT_Q_NOT_DEFINED        99

//  For square dance routines, dancer's BT queue.
#define LDRS_RT_BT_Q             1
#define LDRS_OPP_BT_Q            2
#define LDRS_LFT_BT_Q            3

//  For line dance routines, dancer's BT queue.
//  Remember, the leader knows the routine, and thus whether the other 3 dancers
//  are 1st, 2nd, 3rd in line, or 2nd, 3rd, and 4th.  Note that this must be
//  correctly set when initializing the array dncr_rtine_info[].
#define FRST_IN_LINE_BT_Q             1
#define SCND_IN_LINE_BT_Q             2
#define THRD_IN_LINE_BT_Q             3
#define FRTH_IN_LINE_BT_Q             4

#define DO_IT_SEND_TIME             100

#define STD_DELAY_A                  50
#define STD_DELAY_B                 100
#define STD_DELAY_C                 150
#define STD_DELAY_D                 200
#define STD_DELAY_E                 250
#define STD_DELAY_F                1300
#define STD_DELAY_G                1450


#define LDR_TONE_DELAY              500
#define DNCR_01_TONE_DELAY          928
#define DNCR_02_TONE_DELAY          674
#define DNCR_03_TONE_DELAY          420

#define TONE_FREQ_A               0x200
#define TONE_FREQ_B               0x350
#define TONE_FREQ_C               0x700
#define TONE_FREQ_D              0x1200

#define TONE_TEST_DURATION         3000

//  BlueTooth message stubs
//  BlueTooth messages will be unsigned long (after those sent during initial
//  setup of the BT channel.)
//  All messages will have a message ID (serial number), that are the last 2
//  hex digits, i.e. 0xmmmmmmid.  So unique IDs are 0x00 thru 0xff.  IDs will
//  monotonically increase, and roll over from 255 to 0.  IDs are on a per
//  queue basis.  The LDR and Sprt NXTs will initiallize last_good_rcvd_msg_id
//  for all queues to 0xff, and next_msg_to_send_id to 0x00.
//
//  Initiated messages, usually commands from the LDR, will begin with 0x11 thru
//  0x3f.  The next 4 hex digits are for data.  Thus there are 31 unique
//  commands plus many variations, using the 4 data bits.
//
//  If the leader needs to resend a command, it will increment the first hex
//  digit by 0x4, so the commands become 0x51 to 0x7f.  The last msg ID will be
//  reused.
//
//  If the leader doesn't want to resend a command, but wants to move on to
//  something new, it will increment the first hex digit by 0x8.  The next msg
//  ID will be used.  Upon a message of this type, i.e. beginning with 0x91
//  thru 0xbf, the receiver will not NACK due to msg ID not being exactly one
//  greater than last_good_rcvd_msg_id.
//
//  Upon receiving a message, the NXT will do a sanity check of the command
//  and data, and the msg ID.  If all okay, the reciving NXT will send an ACK
//  message of the form 0xc0mmmmid, where the id echos that of the message just
//  received.  The 4 data digits may be 0, or may answer a question posed in the
//  message.  If the receiving NXT has more data to return than will fit in
//  the 4 data digits, it will use ACK_STUB_WITH_CONTIUATION_1, and send back
//  2 ACK messages for the one command.
//
//  If the receiver thinks there is an error, it will send a NACK message of
//  the form 0xd0mmmmid, where the id echos that of the message just received.
//  The 4 data digits give some indication of the perceived error.
//
//  Protocol is that the LDR always sensds commands or questions.  The sprt
//  dancers always send an ACK or NACK.  When the sprt dancer needs to send
//  info to the LDR, it uses the 4 data digits in ACK response.  If there is
//  more data than can fit in 4 digits, then it uses the
//  ACK_STUB_WITH_CONTIUATION, starting at the highest number, such that the
//  last one sent is the plain ACK_STUB 0xc0000000.
//
#define  ACK_STUB                        0xc0000000
#define  ACK_STUB_WITH_CONTIUATION_1     0xc1000000
#define  ACK_STUB_WITH_CONTIUATION_2     0xc2000000
#define  ACK_STUB_WITH_CONTIUATION_3     0xc3000000
#define  ACK_STUB_WITH_CONTIUATION_F     0xcf000000
#define  NACK_STUB                       0xd0000000
#define  NACK_STUB_WITH_CONTIUATION_1    0xd1000000
#define  NACK_STUB_WITH_CONTIUATION_2    0xd2000000

#define ACK_NACK_MASK                    0xff000000  //  The ack, nack and command
#define COMMAND_MASK                     0xff000000  //  masks are currently equal,
                                                     //  but may not be in the future.
#define MSG_ID_MASK                      0x000000ff
#define QUEUE_MASK                       0x0000ff00
#define SPRT_REQUEST_MASK                0x00ff0000
#define DATA_DIGITS_MASK                 0x00ffff00

#define  RESEND_OFFSET                0x40000000
#define  CONTINUE_FROM_HERE_OFFSET    0x80000000

#define  IM_POLLING_U_STUB            0x21000000
#define  NEXT_QUEUE_TO_USE_STUB       0x22000000  //  0x22wwyyid, ww = bt_chan,
                                                  //  yy = next queue.
#define  PLAY_TONE_NOW_STUB           0x23000000  //  data is freq.
#define  BEGIN_LF_NOW_STUB            0x24000000  //  Do not ACK this, but begin
                                                  //  LF after your standard
                                                  //  delay, that allows all
                                                  //  other dancers to receive
                                                  //  this command.  LF == line
                                                  //  follow.
#define  DO_DANCE_STEP_NOW_STUB       0x25000000  //  data_digits are step.

#define  QUESTION_STUB                0x33000000
#define  END_AND_DISPLAY_STUB         0x3f000000  //  Stop BT comms, display
                                                  //  results.  Used for
                                                  //  debugging.

//  Data stubs
#define  MSG_ID_ERROR                 0x00c10000  //  To be used with NACK response.

#define  WHERE_ARE_YOU                0x00200000  //  data for QUESTION_STUB
#define  MARKER_BLUE_01               0x00212100  //  responses for WHERE_ARE_YOU
#define  MARKER_BLUE_02               0x00212200
#define  MARKER_BLUE_03               0x00212300
#define  MARKER_RED_01                0x00215100
#define  IM_AAW_ALIB                  0x0021ee00  //  I'm alive and well and
                                                  //  living in Billund.

#define  I_NEED                       0x00310000
#define  WHAT_DO_YOU_NEED             0x00300000  //  data for QUESTION_STUB
#define  BRING_BLUE_BRICK             0x00322100  //  responses for DO_YOU_NEED
#define  BRING_GREEN_BRICK            0x00323100
#define  BRING_2X4_BRICK              0x00352400
#define  BRING_2X8_DUPLO              0x00372800
#define  BRING_COFFEE                 0x003e3100  //  i.e. fresh batteries.
#define  BRING_CHOCOLATE              0x003e3200

#define  BRING_LAWYERS                0x003ff100
#define  BRING_GUNS                   0x003ff200
#define  BRING_MONEY                  0x003ff300
#define  BRING_MORE_MONEY             0x003ff400


#define  IM_LOST_IN_RTNE              0x00410000  //  Dancer has lost place in
                                                  //  dance routine, and requesting
                                                  //  help from LDR.
#define  TWIRL_CW                     0x00411100
#define  TWIRL_CCW                    0x00411200


struct  dncr_info
{
    string          dncr_name;
    string          rcx_name;
    unsigned char   dncr_pri_role;
    unsigned char   dncr_sec_role;
    unsigned char   dncr_tert_role;
//    unsigned char   dncr_xger_id;
    unsigned char   dncr_bt_conn;
    unsigned char   dncr_pri_bt_queue;
    unsigned char   dncr_sec_bt_queue;
    unsigned char   dncr_tert_bt_queue;
    bool            bt_connected;
    unsigned char   bt_connection_attempt;
    unsigned char   current_active_queue;
    unsigned short  dncr_tone_freq;
    unsigned long   dncr_std_delay;  //  in milliseconds
};


//  *******************************
//  Global variables
//  *******************************
dncr_info  dncr_rtine_info[5];

unsigned char   my_dncr_id;

unsigned char bt_conn_to_dncr_id[5];
unsigned char dncr_pri_role_to_dncr_id[MAX_VALUE_FOR_ROLE];

//  The following variables are arrays, msg ID is on a per connection and
// a per queue basis.
unsigned char    last_good_rcvd_msg_id[4][10];
unsigned char    next_msg_to_send_id[4][10];


bool display_bt_dbg_data;



//  *******************************
//  Function Prototypes
//  *******************************
bool BTConnection(byte conn, const string name);
bool BTConnectNXT (const string name, byte channel);
bool check_bt_status(char bt_conn);

unsigned char bt_send_remote_number(unsigned char bt_conn,
                                    unsigned char bt_queue,
                                    unsigned long tbs_data);

unsigned char bt_send_response_number(unsigned char bt_conn,
                                      unsigned char bt_queue,
                                      unsigned long tbs_data);

unsigned char bt_receive_remote_number(unsigned char bt_conn,
                                       unsigned char bt_queue,
                                       bool clear_queue,
                                       unsigned long tbr_data_addr);

bool snd_remote_number(unsigned char bt_conn,
                       unsigned char bt_queue,
                       unsigned long tbs_msg);

void snd_number_now(unsigned char bt_conn,
                    unsigned char bt_queue,
                    unsigned long tbs_msg);

bool snd_rcv_by_ref_remote_number(unsigned char bt_conn,
                                  unsigned char bt_queue,
                                  unsigned long tbs_msg,
                                  unsigned long max_response_wait,
                                  unsigned long & actual_response_wait,
                                  unsigned long & response_msg);

bool snd_rcv_by_iom_remote_number(unsigned char bt_conn,
                                  unsigned char bt_queue,
                                  unsigned long tbs_msg,
                                  unsigned long max_response_wait,
                                  unsigned long actual_response_wait_addr,
                                  unsigned long response_msg_addr);

bool rcv_by_ref_remote_number(unsigned char bt_conn,
                              unsigned char bt_queue,
                              unsigned long max_rcv_wait,
                              unsigned long & actual_rcv_wait,
                              unsigned long & received_msg);

bool send_response_msg(unsigned char bt_conn,
                       unsigned char bt_queue,
                       unsigned long tbs_msg);

void pathfind_bt_init();
void wfcr_init();
void ncpr_init();
void wnyc_init();
void wlrn_init();

bool ldr_duo_init_bt_comms(unsigned char req_conn);
void ldr_quad_init_bt_comms();

bool ldr_dncr_init_bt_comms(unsigned char bt_conn);
bool ldr_dncr_confirm_bt_comms(unsigned char bt_conn);

bool sprt_init_bt_comms();
bool fs_init_bt_comms();
char sprt_init_recv_on_q_x(char queue_to_use);
char sprt_init_send_on_q_x(char queue_to_use, unsigned short tbs_data);

void handle_dncr_request(unsigned char dncr_id, unsigned long request_msg);


#endif  //  End of  -  #ifndef _PATHFIND_BT_33_GLOBALS_H

//  End of  -  File Pathfind_BT_33_globals.h

